"""Module for measures of kurtosis."""

import numpy as np
from scipy import special, stats  # type: ignore[import-untyped]


def l_kurt(x: np.ndarray) -> float:
    """Calculate standardized linear kurtosis.

    This measure is a 4th linear moment, which is an
    alternative to conventional moments.

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    lkr : float
        The value of L-Kurtosis.

    References
    ----------
    Hosking, J. R. M. (1990).
    L-moments: analysis and estimation of distributions
    using linear combinations of order statistics.
    Journal of the Royal Statistical Society, Series B. 52 (1): 105-124.
    """
    _x = np.sort(x)
    n = len(_x)
    common = 1 / special.comb(n - 1, (0, 1, 2, 4)) / n
    betas = [
        common[i] * np.nansum(special.comb(np.arange(i, n), i) * _x[i:])
        for i in range(4)
    ]
    l4 = 20 * betas[3] - 30 * betas[2] + 12 * betas[1] - betas[0]
    l2 = 2 * betas[1] - betas[0]
    return float(l4 / l2)


def moors_kurt(x: np.ndarray) -> float:
    """Calculate Moor's vision of kurtosis, based on Z score.

    The kurtosis can now be seen as a measure of the dispersion of
    squared Z around its expectation.
    Alternatively it can be seen to be a measure of the dispersion
    of Z around +1 and -1.

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    mk : float
        The value of Moor's kurtosis.

    References
    ----------
    Moors, J. J. A. (1986).
    The meaning of kurtosis: Darlington reexamined.
    The American Statistician, 40 (4): 283-284,
    """
    return float(np.nanvar(stats.zscore(x, nan_policy="omit") ** 2) + 1)


def moors_octile_kurt(x: np.ndarray) -> float:
    """Calculate Moors measure of kurtosis based on octiles (uncentered, unscaled).

    This measure should be more robust than moment based kurtosis.

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    mok : float
        The value of Moor's octile kurtosis.

    References
    ----------
    Moors, J. J. A. (1988).
    A quantile alternative for kurtosis.
    Journal of the Royal Statistical Society. Series D, 37(1):25-32.
    """
    o1, o2, o3, o5, o6, o7 = np.nanquantile(x, [0.125, 0.25, 0.375, 0.625, 0.75, 0.875])
    return float(((o7 - o5) + (o3 - o1)) / (o6 - o2))


def hogg_kurt(x: np.ndarray) -> float:
    """Calculatie Hogg's kurtosis coefficient.

    It is based on means of values between different percentiles (uncentered, unscaled).
    This measure should be more robust than moment based kurtosis.

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    hgc : float
        The value of Hogg's kurtosis coefficient.

    References
    ----------
    Hogg, R. V. (1972).
    More light on the kurtosis and related statistics.
    Journal of the American Statistical Association, 67(338):422-424.
    """
    p05, p50, p95 = np.nanquantile(x, [0.05, 0.5, 0.95])
    masked_p95 = np.where(x >= p95, x, np.nan)
    masked_p05 = np.where(x <= p05, x, np.nan)
    masked_p50g = np.where(x >= p50, x, np.nan)
    masked_p50l = np.where(x <= p50, x, np.nan)
    return float(
        (np.nanmean(masked_p95) - np.nanmean(masked_p05))
        / (np.nanmean(masked_p50g) - np.nanmean(masked_p50l))
    )


def crow_siddiqui_kurt(x: np.ndarray) -> float:
    """Calculate Crow & Siddiqui kurtosis coefficient.

    It is based on quartiles and percentiles (uncentered, unscaled) and
    tries to compare two different measures of dispersion of the same
    sample.
    This measure should be more robust than moment based kurtosis.

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    csk : float
        The value of Crow & Siddiqui kurtosis coefficient.

    References
    ----------
    Crow, E. L.; Siddiqui, M. (1967).
    Robust estimation of location.
    Journal of the American Statistical Association, 62(318):353-389.
    """
    p025, p25, p75, p975 = np.nanquantile(x, [0.025, 0.25, 0.75, 0.975])
    return float((p975 - p025) / (p75 - p25))


def reza_ma_kurt(x: np.ndarray) -> float:
    """Calculatie Reza & Ma kurtosis coefficient.

    It is based on hexadeciles (uncentered, unscaled) and is very
    similar to Moor's octile kurtosis.
    This measure should be more robust than moment based kurtosis.

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    rmk : float
        The value of Reza & Ma kurtosis coefficient.

    References
    ----------
    Reza, M. S.; Ma, J. (2016).
    ICA and PCA integrated feature extraction for classification.
    2016 IEEE 13th International Conference on Signal Processing (ICSP), 1083-1088.
    """
    h1, h7, h9, h15 = np.nanquantile(x, [0.0625, 0.4375, 0.5625, 0.9375])
    return float(((h15 - h9) + (h7 - h1)) / (h15 - h1))


def staudte_kurt(x: np.ndarray) -> float:
    """Calculate Staudte kurtosis coefficient.

    It is based on inter-percentile ranges (uncentered, unscaled) and
    tries to compare two different measures of dispersion of the same
    sample.
    This measure should be more robust than moment based kurtosis.

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    sk : float
        The value of Staudte kurtosis coefficient.

    References
    ----------
    Staudte, R. G. (2017).
    Inference for quantile measures of kurtosis, peakedness, and tail weight.
    Communications in Statistics-Theory and Methods, 46(7), 3148-3163.
    """
    p10, p33, p66, p90 = np.nanquantile(x, [0.1, 1 / 3, 2 / 3, 0.9])
    return float((p90 - p10) / (p66 - p33))


def schmid_trede_peakedness(x: np.ndarray) -> float:
    """Calculate Schmid and Trder measure of peakedness P.

    It is based on inter-percentile ranges (uncentered, unscaled) and
    tries to compare two different measures of dispersion of the same
    sample.
    This measure should be more robust than moment based kurtosis.

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    stp : float
        The value of measure of peakedness P.

    References
    ----------
    Schmid, F.; Trede, M. (2003).
    Simple tests for peakedness, fat tails and leptokurtosis based on quantiles.
    Computational Statistics and Data Analysis, 43, 1-12.
    """
    p125, p25, p75, p875 = np.nanquantile(x, [0.125, 0.25, 0.75, 0.875])
    return float((p875 - p125) / (p75 - p25))



"""Variation module."""

from .variation import (
    avdev,
    b_index,
    gibbs_m1,
    gibbs_m2,
    mcintosh_d,
    mod_vr,
    negative_extropy,
    range_vr,
    renyi_entropy,
)

__all__ = [
    "avdev",
    "b_index",
    "gibbs_m1",
    "gibbs_m2",
    "mcintosh_d",
    "mod_vr",
    "negative_extropy",
    "range_vr",
    "renyi_entropy",
]



"""Module for measures of categorical variations."""

import math

import numpy as np
from scipy import stats  # type: ignore[import-untyped]


def mod_vr(x: np.ndarray) -> float:
    """Calculate Mode Variation Ratio.

    This ratio could be interpreted as the probability of
    category not being the most frequent.

    Low values of Mode VR correspond to lower variation and
    high values to higher variation.

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    modvr : float
        The value of mode variation ratio.

    References
    ----------
    Wilcox, A. R. (1973).
    Indices of Qualitative Variation and Political Measurement.
    The Western Political Quarterly. 26 (2): 325-343.
    """
    cnts = np.unique(x, return_counts=True, equal_nan=True)[1]
    return float(1 - np.max(cnts) / len(x))


def range_vr(x: np.ndarray) -> float:
    """Calculate Range Variation Ratio.

    Ratio of frequencies of the least and the most common categories.
    This ratio is similar to range or peak-to-peak for real values.

    Low values of Range VR correspond to lower variation and
    high values to higher variation.

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    ranvr : float
        The value of range variation ratio.

    References
    ----------
    Wilcox, A. R. (1973).
    Indices of Qualitative Variation and Political Measurement.
    The Western Political Quarterly. 26 (2): 325-343.
    """
    cnts = np.unique(x, return_counts=True, equal_nan=True)[1]
    return float(np.min(cnts) / np.max(cnts))


def gibbs_m1(x: np.ndarray) -> float:
    """Calculate Gibbs M1 Index.

    M1 can be interpreted as one minus the likelihood that a random pair
    of samples will belong to the same category (standardized likelihood of
    a random pair falling in the same category).

    Low values of Gibbs M1 correspond to lower variation and
    high values to higher variation.

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    m1 : float
        The value of Gibbs M1 index.

    References
    ----------
    Gibbs, J. P.; Poston Jr, D. L. (1975).
    The Division of Labor: Conceptualization and Related Measures.
    Social Forces, 53 (3): 468-476.

    See Also
    --------
    Gini's index of mutability;
    Gini Concentration Index;
    Simpson's measure of diversity;
    Bachi's index of linguistic homogeneity;
    Mueller and Schuessler's index of qualitative variation;
    Gibbs and Martin's index of industry diversification;
    Lieberson's index;
    Blau's index in sociology, psychology and management studies;
    Special case of Tsallis entropy (alpha = 2).
    """
    freq = np.unique(x, return_counts=True, equal_nan=True)[1] / len(x)
    return float(1 - np.sum(freq**2))


def gibbs_m2(x: np.ndarray) -> float:
    """Calculate Gibbs M2 Index.

    M2 can be interpreted as the ratio of the variance of
    the multinomial distribution to the variance of a binomial distribution.

    Low values of Gibbs M2 correspond to lower variation and
    high values to higher variation.

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    m2 : float
        The value of Gibbs M2 index.

    References
    ----------
    Gibbs, J. P.; Poston Jr, D. L. (1975).
    The Division of Labor: Conceptualization and Related Measures.
    Social Forces, 53 (3): 468-476.
    """
    freq = np.unique(x, return_counts=True, equal_nan=True)[1] / len(x)
    k = len(freq)
    return float((k / (k - 1)) * (1 - np.sum(freq**2))) if k > 1 else 0.0


def b_index(x: np.ndarray) -> float:
    """Calculate B Index.

    Normalized to 0-1 range geometric mean of probabilities of all categories.

    Low values of B Index correspond to lower variation and
    high values to higher variation.

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    bi : float
        The value of B index.

    References
    ----------
    Wilcox, A. R. (1973).
    Indices of Qualitative Variation and Political Measurement.
    The Western Political Quarterly. 26 (2): 325-343.
    """
    n = len(x)
    freq = np.unique(x, return_counts=True, equal_nan=True)[1] / n
    return float(1 - (1 - (stats.gmean(freq * len(freq) / n)) ** 2) ** 0.5)


def avdev(x: np.ndarray) -> float:
    """Calculate Average Deviation Analogue.

    Normalized to 0-1 range categorical analogue of the mean deviation.

    Low values of AVDev correspond to lower variation and
    high values to higher variation.

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    avd : float
        The value of AVDev.

    References
    ----------
    Wilcox, A. R. (1973).
    Indices of Qualitative Variation and Political Measurement.
    The Western Political Quarterly. 26 (2): 325-343.
    """
    n = len(x)
    freq = np.unique(x, return_counts=True, equal_nan=True)[1] / n
    k = len(freq)
    mean = n / k
    return float(1 - (np.sum(np.abs(freq - mean)) / (2 * mean * max(k - 1, 1))))


def renyi_entropy(x: np.ndarray, alpha: float = 2, *, normalize: bool = False) -> float:
    """Calculate Renyi entropy (bits).

    Rényi entropy is a quantity that generalizes various notions of entropy,
    including Hartley entropy, Shannon entropy, collision entropy, and min-entropy.

    Low values of Rényi entropy correspond to lower variation and
    high values to higher variation.

    Parameters
    ----------
    x : array_like
        Input array.
    alpha : float, default = 2
        Order of the Rényi entropy.
    normalize : bool, default = False
        Whether to normalize the result.

    Returns
    -------
    ren : float
        The value of Rényi entropy.

    References
    ----------
    Rényi, A. (1961).
    On measures of information and entropy.
    Proceedings of the fourth Berkeley Symposium on Mathematics,
    Statistics and Probability 1960. pp. 547-561.
    """
    if alpha < 0:
        msg = "Parameter alpha should be positive!"
        raise ValueError(msg)
    freq = np.unique(x, return_counts=True, equal_nan=True)[1] / len(x)
    n = len(freq)
    if n <= 1:
        # return 0 if an array is constant
        return 0.0
    normalizer = math.log2(n) if normalize else 1.0
    if alpha == 1:
        # return Shannon entropy to avoid division by 0
        return float(-np.sum(freq * np.log2(freq)) / normalizer)
    return float(1 / (1 - alpha) * math.log2(np.sum(freq**alpha)) / normalizer)


def negative_extropy(x: np.ndarray) -> float:
    """Calculate Negative Information Extropy (bits).

    This measure is complementary to entropy.
    This implementation inverses the sign of extropy to align it with
    all other measures.

    Low values of negative extropy correspond to lower variation and
    high values to higher variation.

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    ext : float
        The value of negative extropy.

    References
    ----------
    Lad, F.; Sanfilippo, G.; Agro, G. (2015).
    Extropy: Complementary dual of entropy.
    Statistical Science, 30(1), 40-58.
    """
    freq = np.unique(x, return_counts=True, equal_nan=True)[1] / len(x)
    p_inv = 1.0 - freq
    return float(-np.sum(p_inv * np.log2(p_inv)))


def mcintosh_d(x: np.ndarray) -> float:
    """Calculate McIntosh's D.

    Ranges from 0 to 1, where 0 corresponds to no diversity,
    and 1 to maximum diversity.

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    mid : float
        The value of McIntosh's D.

    References
    ----------
    McIntosh, R. P. (1967).
    An index of diversity and the relation of certain concepts to diversity.
    Ecology, 48(3), 392-404.
    """
    n = len(x)
    counts = np.unique(x, return_counts=True, equal_nan=True)[1]
    return float((n - np.sum(counts**2) ** 0.5) / (n - n**0.5))



"""Dispersion module."""

from .dispersion import (
    coefficient_of_lvariation,
    coefficient_of_range,
    coefficient_of_variation,
    cole_index_of_dispersion,
    fisher_index_of_dispersion,
    gini_mean_difference,
    inter_expectile_range,
    morisita_index_of_dispersion,
    quartile_coefficient_of_dispersion,
    robust_coefficient_of_variation,
    shamos_estimator,
    standard_quantile_absolute_deviation,
    studentized_range,
)

__all__ = [
    "coefficient_of_lvariation",
    "coefficient_of_range",
    "coefficient_of_variation",
    "cole_index_of_dispersion",
    "fisher_index_of_dispersion",
    "gini_mean_difference",
    "inter_expectile_range",
    "morisita_index_of_dispersion",
    "quartile_coefficient_of_dispersion",
    "robust_coefficient_of_variation",
    "shamos_estimator",
    "standard_quantile_absolute_deviation",
    "studentized_range",
]



"""Kurtosis module."""

from .kurtosis import (
    crow_siddiqui_kurt,
    hogg_kurt,
    l_kurt,
    moors_kurt,
    moors_octile_kurt,
    reza_ma_kurt,
    schmid_trede_peakedness,
    staudte_kurt,
)

__all__ = [
    "crow_siddiqui_kurt",
    "hogg_kurt",
    "l_kurt",
    "moors_kurt",
    "moors_octile_kurt",
    "reza_ma_kurt",
    "schmid_trede_peakedness",
    "staudte_kurt",
]



"""Central tendency module."""

from .central_tendency import (
    contraharmonic_mean,
    gastwirth_location,
    grenanders_m,
    half_sample_mode,
    hodges_lehmann_sen_location,
    midhinge,
    midmean,
    midrange,
    standard_trimmed_harrell_davis_quantile,
    tau_location,
    trimean,
)

__all__ = [
    "contraharmonic_mean",
    "gastwirth_location",
    "grenanders_m",
    "half_sample_mode",
    "hodges_lehmann_sen_location",
    "midhinge",
    "midmean",
    "midrange",
    "standard_trimmed_harrell_davis_quantile",
    "tau_location",
    "trimean",
]



"""Module for measures of skewness."""

from __future__ import annotations

import numpy as np
from scipy import integrate, special, stats  # type: ignore[import-untyped]

from obscure_stats.central_tendency import half_sample_mode


def l_skew(x: np.ndarray) -> float:
    """Calculate standardized linear skewness.

    This measure is a 3rd linear moment, which is an
    alternative to conventional moments.

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    lsk : float.
        The value of L-Skewness.

    References
    ----------
    Hosking, J. R. M. (1990).
    L-moments: analysis and estimation of distributions
    using linear combinations of order statistics.
    Journal of the Royal Statistical Society, Series B. 52 (1): 105-124.
    """
    _x = np.sort(x)
    n = len(_x)
    common = 1 / special.comb(n - 1, (0, 1, 2)) / n
    betas = [
        common[i] * np.nansum(special.comb(np.arange(i, n), i) * _x[i:])
        for i in range(3)
    ]
    l3 = 6 * betas[2] - 6 * betas[1] + betas[0]
    l2 = 2 * betas[1] - betas[0]
    return float(l3 / l2)


def pearson_mode_skew(x: np.ndarray) -> float:
    """Calculate Pearson's mode skew coefficient.

    This measure could be unstable due mode instability.

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    pmods : float
        The value of Pearson's mode skew coefficient.

    References
    ----------
    Pearson, E. S.; Hartley, H. O. (1966).
    Biometrika Tables for Statisticians, vols. I and II.
    Cambridge University Press, Cambridge.
    """
    mean = np.nanmean(x)
    mode = stats.mode(x)[0]
    std = np.nanstd(x)
    return float((mean - mode) / std)


def bickel_mode_skew(x: np.ndarray) -> float:
    """Calculate Robust Mode skew with half sample mode.

    This measure should be more stable than Pearson mode
    skewness coefficient.

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    bms : float
        The value of Bickel's mode skew coefficient.

    References
    ----------
    Bickel, D. R. (2002).
    Robust estimators of the mode and skewness of continuous data.
    Computational Statistics & Data Analysis, Elsevier, 39(2), 153-163.
    """
    mode = half_sample_mode(x)
    return float(np.nanmean(np.sign(np.asarray(x) - mode)))


def pearson_median_skew(x: np.ndarray) -> float:
    """Calculatie Pearson's median skew coefficient.

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    pmeds : float
        The value of Pearson's median skew coefficient.

    References
    ----------
    Pearson, E.S.; Hartley, H.O. (1966).
    Biometrika Tables for Statisticians, vols. I and II.
    Cambridge University Press, Cambridge.
    """
    mean = np.nanmean(x)
    median = np.nanmedian(x)
    std = np.nanstd(x)
    return float(3.0 * (mean - median) / std)


def medeen_skew(x: np.ndarray) -> float:
    """Calculate Medeen's skewness statistic.

    This measure is similar to Pearson median skewness coefficient
    but uses different normalization (mean absolute deviation from the
    median).

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    mss : float
        The value of Medeen's skewness statistic.

    References
    ----------
    Groeneveld, R.A.; Meeden, G. (1984).
    Measuring Skewness and Kurtosis.
    The Statistician. 33 (4): 391-399.
    """
    median = np.nanmedian(x)
    mean = np.nanmean(x)
    return float((mean - median) / np.nanmean(np.abs(x - median)))


def bowley_skew(x: np.ndarray) -> float:
    """Calculate Bowley's skewness coefficinet.

    Also known as Yule-Kendall skewness coefficient.
    It is based on quartiles (uncentered, unscaled) and compares the distance
    between the median and each of the two quartiles.
    This measure should be more robust than moment based skewness.

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    bsk : float
        The value of Bowley's skewness coefficinet.

    References
    ----------
    Bowley, A. L. (1901).
    Elements of Statistics.
    P.S. King and Son, London.
    """
    q1, q2, q3 = np.nanquantile(x, [0.25, 0.5, 0.75])
    return float((q3 + q1 - 2 * q2) / (q3 - q1))


def groeneveld_range_skew(x: np.ndarray) -> float:
    """Calculate Groeneveld's skewness coefficinet.

    This measure should be more robust than moment based skewness.
    This is the implementation of 'b4' skewness from the paper.

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    grs : float
        The value of Groeneveld's range skewness coefficinet.

    References
    ----------
    Groeneveld, R. A.; Meeden, G. (1984).
    Measuring Skewness and Kurtosis.
    The Statistician. 33 (4): 391-399.
    """
    m = np.nanmedian(x)
    min_ = np.nanmin(x)
    max_ = np.nanmax(x)
    return float((max_ + min_ - 2 * m) / (max_ - min_))


def kelly_skew(x: np.ndarray) -> float:
    """Calculate Kelly's skewness coefficinet.

    It is based on deciles (uncentered, unscaled).
    Similar to Bowley skewness coefficient, but instead of
    quartiles, deciles are used.
    This measure should be more robust than moment based skewness.

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    ksc : float
        The value of Kelly's skewness coefficinet.

    References
    ----------
    David, F. N.; Johnson, N. L. (1956).
    Some tests of significance with ordered variables.
    J. R. Stat. Soc. Ser. B Stat. Methodol. 18, 1-31.
    """
    d1, d5, d9 = np.nanquantile(x, [0.1, 0.5, 0.9])
    return float((d9 + d1 - 2 * d5) / (d9 - d1))


def hossain_adnan_skew(x: np.ndarray) -> float:
    """Calculate Houssain and Adnan skewness coefficient.

    It is based on differences from the median, and is somewhar similar
    to Bickel mode skewness.
    This measure should be more robust than moment based skewness.

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    has : float
        The value of Houssain and Adnan skewness coefficient.

    References
    ----------
    Hossain, M. F.; Adnan, M. A. S. A (2007).
    A New Approach to Determine the Asymmetry of a Distribution.
    Journal of Applied St atistical Science, Vol.15, pp. 127-134.
    """
    diff = x - np.nanmedian(x)
    return float(np.nanmean(diff) / np.nanmean(np.abs(diff)))


def forhad_shorna_rank_skew(x: np.ndarray) -> float:
    """Calculate Forhad-Shorna coefficient of rank skewness.

    This measure is similar to Houssain and Adnan skewness coefficient,
    but uses differences in ranks instead of absolute differences.
    This measure should be more robust than moment based skewness.
    The array will be flatten before any calculations.

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    fsrs : float
        The value of Forhad-Shorna coefficient of rank skewness.

    References
    ----------
    Shorna, U. S.; Hossain, M. (2019).
    A New Approach to Determine the Coefficient of Skewness and
    An Alternative Form of Boxplot.
    arXiv preprint arXiv:1908.06400.
    """
    _x = np.ravel(x)
    mr = (np.nanmin(_x) + np.nanmax(_x)) * 0.5
    arr = np.r_[_x, mr]
    arr_ranked = stats.rankdata(arr, method="average", nan_policy="omit")
    diff = arr_ranked[-1] - arr_ranked
    diff = diff[:-1]
    return float(np.nansum(diff) / np.nansum(np.abs(diff)))


def auc_skew_gamma(x: np.ndarray, dp: float = 0.01) -> float:
    """Calculate area under the curve of generalized Bowley skewness coefficients.

    This measure tries to combine multiple generalized Bowley skewness coefficients
    into one value.

    Parameters
    ----------
    x : array_like
        Input array.
    dp : float, default = 0.01
        Step used in calculating area under the curve (integrating).

    Returns
    -------
    aucbs : float
        The value of AUC Bowley skewness.

    References
    ----------
    Arachchige, C. N.; & Prendergast, L. A. (2019).
    Mean skewness measures.
    arXiv preprint arXiv:1912.06996.
    """
    n = int(1 / dp)
    half_n = n // 2
    qs = np.nanquantile(x, np.r_[np.linspace(0, 1, n), 0.5])
    med = qs[-1]
    qs = qs[:-1]
    qs_low = qs[:half_n]
    qs_high = qs[-half_n:]
    skews = (qs_low + qs_high - 2 * med) / (qs_high - qs_low)
    return float(integrate.trapezoid(skews, dx=dp))


def left_quantile_weight(x: np.ndarray, q: float = 0.25) -> float:
    """Calculate left quantile weight (LQW).

    It is based on inter-percentile ranges (uncentered, unscaled) of the
    left tail of the distribution.

    Parameters
    ----------
    x : array_like
        Input array.
    q : float, default = 0.25
        Quantile to use for the calculation, (0.0, 0.5)

    Returns
    -------
    lqw : float
        The value of left quantile weight.

    References
    ----------
    Brys, G.; Hubert, M.; Struyf, A. (2006).
    Robust measures of tail weight.
    Computational Statistics and Data Analysis 50(3), 733-759.
    """
    min_q, max_q = 0.0, 0.5
    if q <= min_q or q >= max_q:
        msg = "Parameter q should be in range (0, 0.5)."
        raise ValueError(msg)
    lower_quantile, q025, upper_quantile = np.nanquantile(
        x, [q * 0.5, 0.25, (1 - q) * 0.5]
    )
    return float(
        -(upper_quantile + lower_quantile - 2 * q025)
        / (upper_quantile - lower_quantile)
    )


def right_quantile_weight(x: np.ndarray, q: float = 0.75) -> float:
    """Calculate right quantile weight (RQW).

    It is based on inter-percentile ranges (uncentered, unscaled) of the
    right tail of the distribution.

    Parameters
    ----------
    x : array_like
        Input array.
    q : float, default = 0.75
        Quantile to use for the calculation, (0.5, 1.0).

    Returns
    -------
    rqw : float
        The value of right quantile weight.

    References
    ----------
    Brys, G.; Hubert, M.; Struyf, A. (2006).
    Robust measures of tail weight.
    Computational Statistics and Data Analysis 50(3), 733-759.
    """
    min_q, max_q = 0.5, 1.0
    if q <= min_q or q >= max_q:
        msg = "Parameter q should be in range (0.5, 1.0)."
        raise ValueError(msg)
    lower_quantile, q075, upper_quantile = np.nanquantile(
        x, [1 - q * 0.5, 0.75, (1 + q) * 0.5]
    )
    return float(
        (lower_quantile + upper_quantile - 2 * q075) / (lower_quantile - upper_quantile)
    )



"""Module for measures of dispersion."""

import numpy as np
from scipy import special, stats  # type: ignore[import-untyped]


def studentized_range(x: np.ndarray) -> float:
    """Calculate range normalized by standard deviation.

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    sr : float
        The value of the studentized range.

    References
    ----------
    Student (1927).
    Errors of routine analysis.
    Biometrika. 19 (1/2): 151-164.
    """
    std = np.nanstd(x)
    maximum = np.nanmax(x)
    minimum = np.nanmin(x)
    return float((maximum - minimum) / std)


def coefficient_of_lvariation(x: np.ndarray) -> float:
    """Calculate linear coefficient of variation.

    L-CV is the L-scale (half of mean absolute deviation) divided
    by L-mean (the same as regular mean).

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    lcv : float
        The value of the linear coefficient of variation.

    References
    ----------
    Hosking, J. R. M. (1990).
    L-moments: analysis and estimation of distributions
    using linear combinations of order statistics.
    Journal of the Royal Statistical Society, Series B. 52 (1): 105-124.
    """
    l1 = np.nanmean(x)
    _x = np.sort(x)
    n = len(_x)
    common = 1 / special.comb(n - 1, 1) / n
    beta_1 = common * np.nansum(special.comb(np.arange(1, n), 1) * _x[1:])
    l2 = 2 * beta_1 - l1
    return float(l2 / l1)


def coefficient_of_variation(x: np.ndarray) -> float:
    """Calculate coefficient of variation (Standard deviation / Mean).

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    cv : float
        The value of the coefficient of variation.

    References
    ----------
    Brown, C. E. (1998).
    Coefficient of Variation.
    Applied Multivariate Statistics in Geohydrology and Related Sciences. Springer.
    """
    return float(np.nanstd(x) / np.nanmean(x))


def robust_coefficient_of_variation(x: np.ndarray) -> float:
    """Calculate robust coefficient of variation.

    It is based on median absolute deviation from the median, i.e. median
    absolute deviation from the median divided by the median.

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    rcv : float
        The value of the robust coefficient of variation.

    References
    ----------
    Reimann, C.; Filzmoser; P.; Garrett, R. G.; Dutter, R. (2008).
    Statistical Data Analysis Explained: Applied Environmental Statistics with R.
    John Wiley and Sons, New York.
    """
    med = np.nanmedian(x)
    med_abs_dev = np.nanmedian(np.abs(x - med))
    return float(1.4826 * med_abs_dev / med)


def quartile_coefficient_of_dispersion(x: np.ndarray) -> float:
    """Calculate quartile coefficient of dispersion (IQR / Midhinge).

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    qcd : float
        The value of the quartile coefficient of dispersion.

    References
    ----------
    Shapiro, H. M. (2005).
    Practical flow cytometry.
    John Wiley & Sons.
    """
    q1, q3 = np.nanquantile(x, [0.25, 0.75])
    return float(0.75 * (q3 - q1) / (q3 + q1))


def fisher_index_of_dispersion(x: np.ndarray) -> float:
    """Calculate Fisher's index of dispersion.

    It is very similar to the coefficient of variation but uses unnormalized
    variation instead of the standard deviation.

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    fi : float
        The value of the Fisher's index of dispersion.

    References
    ----------
    Fisher, R. A. (1925).
    Statistical methods for research workers.
    Hafner, New York.
    """
    return float((len(x) - 1) * np.nanvar(x) / np.nanmean(x))


def morisita_index_of_dispersion(x: np.ndarray) -> float:
    """Calculate Morisita's index of dispersion.

    Morisita's index of dispersion is the scaled probability that two
    points chosen at random from the whole population are in the same sample.

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    mi : float
        The value of the Morisita's index.

    References
    ----------
    Morisita, M. (1959).
    Measuring the dispersion and the analysis of distribution patterns.
    Memoirs of the Faculty of Science, Kyushu University Series e. Biol. 2: 215-235
    """
    x_sum = np.nansum(x)
    return float(len(x) * (np.nansum(np.square(x)) - x_sum) / (x_sum**2 - x_sum))


def standard_quantile_absolute_deviation(x: np.ndarray) -> float:
    """Calculate standard quantile absolute deviation.

    This measure is a robust measure of dispersion, that has higher
    gaussian efficiency, but lower breaking point than MAD.

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    sqad : float
        The value of the standard quantile absolute deviation.

    References
    ----------
    Akinshin, A. (2022).
    Quantile absolute deviation.
    arXiv preprint arXiv:2208.13459.
    """
    med = np.nanmedian(x)
    n = len(x)
    # finite sample correction
    k = 1.0 + 0.762 / n + 0.967 / n**2
    # constant value that maximizes efficiency for normal distribution
    q = 0.6826894921370850  # stats.norm.cdf(1) - stats.norm.cdf(-1)
    return float(k * np.nanquantile(np.abs(x - med), q=q))


def shamos_estimator(x: np.ndarray) -> float:
    """Calculate Shamos robust estimator of dispersion.

    This measure is complementary to Hodges-Lehmann-Sen estimator.

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    se : float
        The value of Shamos estimator.

    References
    ----------
    Shamos, M. I. (1976).
    Geometry and statistics: Problems at the interface (p. 0032).
    Carnegie-Mellon University. Department of Computer Science.

    Notes
    -----
    This implementation uses cartesian product, so the time and memory complexity
    are N^2. It is best to not use it on large arrays.

    See Also
    --------
    obscure_stats.central_tendency.hodges_lehmann_sen_location - Hodges-Lehmann-Sen loc.
    """
    # In the original paper authors suggest use only upper triangular
    # of the cartesian product, but in this implementation we use
    # whole matrix, which is equvalent.
    product = np.meshgrid(x, x, sparse=True)
    return float(np.nanmedian(np.abs(product[0] - product[1])))


def coefficient_of_range(x: np.ndarray) -> float:
    """Calculate coefficient of range (Range / Midrange).

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    cr : float
        The value of the range coefficient.

    References
    ----------
    Yadav, S. K., Singh, S.,  &  Gupta, R. (2019).
    Measures of Dispersion.
    In Biomedical Statistics (pp. 59-70). Springer, Singapore
    """
    min_ = np.nanmin(x)
    max_ = np.nanmax(x)
    return float((max_ - min_) / (max_ + min_))


def cole_index_of_dispersion(x: np.ndarray) -> float:
    """Calculate Cole's index of dispersion.

    Higher values mean higher dispersion.
    This measure is invariant to multiplication.

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    ci : float
        The value of the Cole's index of dispersion.

    References
    ----------
    Cole, L. C. (1946).
    A theory for analyzing contagiously distributed populations.
    Ecology. 27 (4): 329-341.
    """
    return float(np.nansum(np.square(x)) / np.nansum(x) ** 2)


def gini_mean_difference(x: np.ndarray) -> float:
    """Calculate Gini Mean Difference.

    Alternative measure of variability to the usual standard deviation.

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    gmd : float
        The value of the Gini Mean Difference.

    References
    ----------
    Yitzhaki, S.; Schechtman, E. (2013).
    The Gini Methodology.
    Springer, New York.

    Notes
    -----
    This implementation uses cartesian product, so the time and memory complexity
    are N^2. It is best to not use it on large arrays.
    """
    n = len(x)
    product = np.meshgrid(x, x, sparse=True)
    return float(np.nansum(np.abs(product[0] - product[1])) / (n * (n - 1)))


def inter_expectile_range(x: np.ndarray) -> float:
    """Calculate inter expectile range (IER).

    It is the same as IQR, but uses expectile instead of quantile.

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    ier : float
        The value of the inter expectile range.

    References
    ----------
    Bellini, F.; Mercuri, L.; Rroji, E. (2018)
    Implicit expectiles and measures of implied volatility.
    Quantitative Finance, 18(11), pp. 1851-1864.
    """
    _x = np.ravel(x)
    _x = _x[np.isfinite(_x)]
    if len(_x) <= 1:
        return np.nan
    return float(stats.expectile(_x, 0.75) - stats.expectile(_x, 0.25))



"""Association module."""

from .association import (
    blomqvist_beta,
    chatterjee_xi,
    concordance_correlation,
    concordance_rate,
    fechner_correlation,
    gaussain_rank_correlation,
    normalized_chatterjee_xi,
    quantile_correlation,
    rank_minrelation_coefficient,
    symmetric_chatterjee_xi,
    tanimoto_similarity,
    tukey_correlation,
    winsorized_correlation,
    zhang_i,
)

__all__ = [
    "blomqvist_beta",
    "chatterjee_xi",
    "concordance_correlation",
    "concordance_rate",
    "fechner_correlation",
    "gaussain_rank_correlation",
    "normalized_chatterjee_xi",
    "quantile_correlation",
    "rank_minrelation_coefficient",
    "symmetric_chatterjee_xi",
    "tanimoto_similarity",
    "tukey_correlation",
    "winsorized_correlation",
    "zhang_i",
]



"""Skewness module."""

from .skewness import (
    auc_skew_gamma,
    bickel_mode_skew,
    bowley_skew,
    forhad_shorna_rank_skew,
    groeneveld_range_skew,
    hossain_adnan_skew,
    kelly_skew,
    l_skew,
    left_quantile_weight,
    medeen_skew,
    pearson_median_skew,
    pearson_mode_skew,
    right_quantile_weight,
)

__all__ = [
    "auc_skew_gamma",
    "bickel_mode_skew",
    "bowley_skew",
    "forhad_shorna_rank_skew",
    "groeneveld_range_skew",
    "hossain_adnan_skew",
    "kelly_skew",
    "l_skew",
    "left_quantile_weight",
    "medeen_skew",
    "pearson_median_skew",
    "pearson_mode_skew",
    "right_quantile_weight",
]



"""Module for association measures."""

from __future__ import annotations

import warnings

import numpy as np
from scipy import stats  # type: ignore[import-untyped]

from obscure_stats.dispersion import gini_mean_difference


def _check_arrays(x: np.ndarray, y: np.ndarray) -> bool:
    """Check arrays.

    - Equal lenghts of the arrays;
    - Enough lenghts of the arrays;
    - Constant input;
    - Contains inf.
    """
    _x = np.ravel(x)
    _y = np.ravel(y)
    if len(_x) != len(_y):
        warnings.warn(
            "Lenghts of the inputs do not match, please check the arrays.", stacklevel=2
        )
        return True
    if len(_x) <= 1:
        warnings.warn(
            "Lenghts of the inputs are too small, please check the arrays.",
            stacklevel=2,
        )
        return True
    if all(np.isclose(_x, np.nanmin(_x), equal_nan=False)) or all(
        np.isclose(_y, np.nanmin(_y), equal_nan=False)
    ):
        warnings.warn(
            "One of the input arrays is constant;"
            " the correlation coefficient is not defined.",
            stacklevel=2,
        )
        return True
    if any(np.isinf(_x)) or any(np.isinf(_y)):
        warnings.warn(
            "One of the input arrays contains inf, please check the array.",
            stacklevel=2,
        )
        return True
    if (np.isnan(_x).sum() >= len(_x) - 1) or (np.isnan(_y).sum() >= len(_y) - 1):
        warnings.warn(
            "One of the input arrays has too many missing values,"
            " please check the arrays.",
            stacklevel=2,
        )
        return True
    return False


def _prep_arrays(x: np.ndarray, y: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    """Prepare data for downstream task."""
    _x = np.ravel(x)
    _y = np.ravel(y)
    notnan = np.isfinite(_x) & np.isfinite(_y)
    _x = _x[notnan]
    _y = _y[notnan]
    return _x, _y


def chatterjee_xi(x: np.ndarray, y: np.ndarray) -> float:
    """Calculate Xi correlation coefficient.

    Another variation of rank correlation which does not make any assumptions about
    underlying distributions of the variable.

    It ranges from 0 (variables are completely independent) to 1
    (one is a measurable function of the other). But a lot of the times the maximum
    value of the coefficient is lower than 1.

    This implementation does not break ties at random, instead
    it break ties depending on order. This makes it dependent on
    data sorting, which could be useful in application like time
    series.

    The arrays will be flatten before any calculations.

    Parameters
    ----------
    x : array_like
        Input array.
    y : array_like
        Input array.

    Returns
    -------
    xi : float.
        The value of the xi correlation coefficient.

    References
    ----------
    Chatterjee, S. (2021).
    A new coefficient of correlation.
    Journal of the American Statistical Association, 116(536), 2009-2022.

    Notes
    -----
    This measure is assymetric: (x, y) != (y, x).
    """
    if _check_arrays(x, y):
        return np.nan
    x, y = _prep_arrays(x, y)
    if _check_arrays(x, y):
        return np.nan
    # heavily inspired by https://github.com/czbiohub-sf/xicor/issues/17#issue-965635013
    n = len(x)
    y_forward_ordered = y[np.argsort(x)]
    _, y_unique_indexes, y_counts = np.unique(
        y_forward_ordered, return_inverse=True, return_counts=True
    )
    right = np.cumsum(y_counts)[y_unique_indexes]
    left = np.cumsum(y_counts[::-1])[len(y_counts) - y_unique_indexes - 1]
    return float(
        1.0 - 0.5 * np.sum(np.abs(np.diff(right))) / np.mean(left * (n - left))
    )


def concordance_correlation(x: np.ndarray, y: np.ndarray) -> float:
    """Calculate concordance correlation coefficient.

    The main difference between Pearson's R and CCC is that CCC
    takes bias between variables into account.

    CCC measures the agreement between two variables, e.g.,
    to evaluate reproducibility or for inter-rater reliability.

    The arrays will be flatten before any calculations.

    Parameters
    ----------
    x : array_like
        Input array.
    y : array_like
        Input array.

    Returns
    -------
    ccc : float.
        The value of the concordance correlation coefficient.

    References
    ----------
    Lin, L. I. (1989).
    A concordance correlation coefficient to evaluate reproducibility.
    Biometrics. 45 (1): 255-268.
    """
    if _check_arrays(x, y):
        return np.nan
    x, y = _prep_arrays(x, y)
    if _check_arrays(x, y):
        return np.nan
    std_x = np.std(x, ddof=0)
    std_y = np.std(y, ddof=0)
    w = std_y / std_x
    v = (np.mean(x) - np.mean(y)) ** 2 / (std_x * std_y) ** 0.5
    x_a = 2 / (v**2 + w + 1 / w)
    p = np.corrcoef(x, y)[0][1]
    return float(p * x_a)


def concordance_rate(x: np.ndarray, y: np.ndarray) -> float:
    """Calculate conventional concordance rate.

    Also known as quadrant count ratio.
    It could be seen as simplified version of Pearson's R.

    It differs from quadrant count ratio by adding and exclusion zone
    variation has an option for an exclusion zone. It is based on the
    standard error of the mean and will exlucde points that are in the
    range of mean+-sem.

    The arrays will be flatten before any calculations.

    Parameters
    ----------
    x : array_like
        Input array.
    y : array_like
        Input array.

    Returns
    -------
    cr : float.
        The value of the concordance rate.

    References
    ----------
    Holmes, P. (2001).
    Correlation: From Picture to Formula.
    Teaching Statistics. 23 (3): 67-71.

    See Also
    --------
    Quadrant count ratio.
    """
    if _check_arrays(x, y):
        return np.nan
    x, y = _prep_arrays(x, y)
    if _check_arrays(x, y):
        return np.nan
    n = len(x)
    mean_x = np.sum(x) / n
    mean_y = np.sum(y) / n
    sem_x = np.std(x, ddof=0) / n**0.5
    sem_y = np.std(y, ddof=0) / n**0.5
    return float(
        (
            np.sum((x >= mean_x + sem_x) & (y >= mean_y + sem_y))
            - np.sum((x <= mean_x - sem_x) & (y >= mean_y + sem_y))
            + np.sum((x <= mean_x - sem_x) & (y <= mean_y - sem_y))
            - np.sum((x >= mean_x + sem_x) & (y <= mean_y - sem_y))
        )
        / n
    )


def symmetric_chatterjee_xi(x: np.ndarray, y: np.ndarray) -> float:
    """Calculate symmetric Xi correlation coefficient.

    Another variation of rank correlation which does not make any assumptions about
    underlying distributions of the variable.

    It ranges from 0 (variables are completely independent) to 1
    (one is a measurable function of the other). But a lot of the times the maximum
    value of the coefficient is lower than 1.

    This implementation does not break ties at random, instead
    it break ties depending on order. This makes it dependent on
    data sorting, which could be useful in application like time
    series.

    The arrays will be flatten before any calculations.

    Parameters
    ----------
    x : array_like
        Input array.
    y : array_like
        Input array.

    Returns
    -------
    sxi : float.
        The value of the symmetric xi correlation coefficient.

    References
    ----------
    Chatterjee, S. (2021).
    A new coefficient of correlation.
    Journal of the American Statistical Association, 116(536), 2009-2022.

    See Also
    --------
    obscure_stats.associaton.chatterjee_xi - Chatterjee Xi coefficient.
    """
    if _check_arrays(x, y):
        return np.nan
    x, y = _prep_arrays(x, y)
    if _check_arrays(x, y):
        return np.nan
    n = len(x)
    # y ~ f(x)
    y_forward_ordered = y[np.argsort(x)]
    _, y_unique_indexes, y_counts = np.unique(
        y_forward_ordered, return_inverse=True, return_counts=True
    )
    right_xy = np.cumsum(y_counts)[y_unique_indexes]
    left_xy = np.cumsum(y_counts[::-1])[len(y_counts) - y_unique_indexes - 1]
    # x ~ f(y)
    x_forward_ordered = x[np.argsort(y)]
    _, x_unique_indexes, x_counts = np.unique(
        x_forward_ordered, return_inverse=True, return_counts=True
    )
    right_yx = np.cumsum(x_counts)[x_unique_indexes]
    left_yx = np.cumsum(x_counts[::-1])[len(x_counts) - x_unique_indexes - 1]
    # choose the highest from the two
    return float(
        1.0
        - min(
            0.5 * np.sum(np.abs(np.diff(right_xy))) / np.mean(left_xy * (n - left_xy)),
            0.5 * np.sum(np.abs(np.diff(right_yx))) / np.mean(left_yx * (n - left_yx)),
        )
    )


def zhang_i(x: np.ndarray, y: np.ndarray) -> float:
    """Calculate I correlation coefficient proposed by Q. Zhang.

    This coefficient combines Spearman and Chatterjee rank correlation coefficients
    to get higher sensetivity to complex nonlinear relationships between variables.

    The arrays will be flatten before any calculations.

    Parameters
    ----------
    x : array_like
        Input array.
    y : array_like
        Input array.

    Returns
    -------
    zi : float.
        The value of the Zhang I.

    References
    ----------
    Zhang, Q. (2025).
    On the extensions of the Chatterjee-Spearman test.
    Journal of Nonparametric Statistics, 1-30.

    See Also
    --------
    scipy.stats.spearmanr - Spearman R coefficient.
    obscure_stats.associaton.symmetric_chatterjee_xi - Chatterjee Xi coefficient.
    """
    if _check_arrays(x, y):
        return np.nan
    x, y = _prep_arrays(x, y)
    if _check_arrays(x, y):
        return np.nan
    return float(
        min(
            1.0,
            max(
                abs(stats.spearmanr(x, y, nan_policy="omit")[0]),
                2.5**0.5 * symmetric_chatterjee_xi(x, y),
            ),
        )
    )


def tanimoto_similarity(x: np.ndarray, y: np.ndarray) -> float:
    """Calculate Tanimoto similarity.

    It is very similar to Jaccard or Cosine similarity but differs in how
    dot product is normalized.
    This version is designed for numeric values, instead of sets.

    The arrays will be flatten before any calculations.

    Parameters
    ----------
    x : array_like
        Input array.
    y : array_like
        Input array.

    Returns
    -------
    ts : float.
        The value of the Tanimoto similarity measure

    References
    ----------
    Rogers, D. J.; Tanimoto, T. T. (1960).
    A Computer Program for Classifying Plants.
    Science. 132 (3434): 1115-8.

    See Also
    --------
    Jaccard similarity
    Cosine similarity
    """
    if _check_arrays(x, y):
        return np.nan
    x, y = _prep_arrays(x, y)
    if _check_arrays(x, y):
        return np.nan
    xy = np.mean(x * y)
    xx = np.mean(x**2)
    yy = np.mean(y**2)
    return float(xy / (xx + yy - xy))


def blomqvist_beta(x: np.ndarray, y: np.ndarray) -> float:
    """Calculate Blomqvist's beta.

    Also known as medial correlation. It is similar to Spearman Rho
    and Kendall Tau correlations, but have some advantages over them.

    The arrays will be flatten before any calculations.

    Parameters
    ----------
    x : array_like
        Input array.
    y : array_like
        Input array.

    Returns
    -------
    bb : float.
        The value of the Blomqvist's beta.

    References
    ----------
    Blomqvist,  N. (1950).
    On a measure of dependence between two random variables.
    Annals of Mathematical Statistics, 21, 593-600.

    Schmid, F.; Schmidt, R. (2007).
    Nonparametric Inference on Multivariate Versions of
    Blomqvist's Beta and Related Measures of Tail Dependence.
    Metrika, 66(3), 323-354.

    See Also
    --------
    scipy.stats.spearmanr - Spearman R coefficient.
    scipy.stats.kendalltau - Kendall Tau coefficient.
    """
    if _check_arrays(x, y):
        return np.nan
    x, y = _prep_arrays(x, y)
    if _check_arrays(x, y):
        return np.nan
    med_x = np.median(x)
    med_y = np.median(y)
    return float(np.mean(np.sign((x - med_x) * (y - med_y))))


def fechner_correlation(x: np.ndarray, y: np.ndarray) -> float:
    """Calculate Fechner correlation.

    It is similar to Bloomqvist beta and Pearson correlation.

    The arrays will be flatten before any calculations.

    Parameters
    ----------
    x : array_like
        Input array.
    y : array_like
        Input array.

    Returns
    -------
    fc : float.
        The value of the Fechner correlation.

    References
    ----------
    Latyshev, A.V.; Koldanov, P.A. (2014).
    Investigation of Connections Between Pearson and Fechner Correlations
    in Market Network: Experimental Study.
    Models, Algorithms, and Technologies for Network Analysis, 156, 175-182

    See Also
    --------
    obscure_stats.associaton.blomqvist_beta - Bloomqvist beta.
    """
    if _check_arrays(x, y):
        return np.nan
    x, y = _prep_arrays(x, y)
    if _check_arrays(x, y):
        return np.nan
    avg_x = np.mean(x)
    avg_y = np.mean(y)
    return float(np.mean(np.sign(x - avg_x) * np.sign(y - avg_y)))


def winsorized_correlation(x: np.ndarray, y: np.ndarray, k: float = 0.1) -> float:
    """Calculate winsorized correlation coefficient.

    This correlation is a robust alternative of the Pearson correlation.

    The arrays will be flatten before any calculations.

    Parameters
    ----------
    x : array_like
        Input array.
    y : array_like
        Input array.
    k : float, default = 0.1
        The percentages of values to winsorize on each side of the arrays.

    Returns
    -------
    wcr : float.
        The value of the winsorized correlation.

    References
    ----------
    Wilcox, R. R. (1993).
    Some Results on a Winsorized Correlation Coefficient.
    British Journal of Mathematical and Statistical Psychology, 46, 339-349.

    See Also
    --------
    scipy.stats.pearsonr - Pearson correlation coefficient.
    """
    if _check_arrays(x, y):
        return np.nan
    x, y = _prep_arrays(x, y)
    if _check_arrays(x, y):
        return np.nan
    x_w = stats.mstats.winsorize(x, (k, k))
    y_w = stats.mstats.winsorize(y, (k, k))
    return float(np.corrcoef(x_w, y_w)[0, 1])


def rank_minrelation_coefficient(x: np.ndarray, y: np.ndarray) -> float:
    """Calculate rank minrelation coefficient.

    This measure estimates p(y > x) when x and y are continuous random variables.
    In short, if a variable x exhibits a minrelation to y then,
    as x increases, y is likely to increases too.

    The arrays will be flatten before any calculations.

    Parameters
    ----------
    x : array_like
        Input array.
    y : array_like
        Input array.

    Returns
    -------
    rmc : float.
        The value of the rank minrelation coefficient.

    References
    ----------
    Meyer, P. E. (2013).
    A Rank Minrelation-Majrelation Coefficient.
    arXiv preprint arXiv:1305.2038.

    Notes
    -----
    This measure is assymetric: (x, y) != (y, x).

    See Also
    --------
    Concordance rate.
    Concordance correlation coefficient.
    """
    if _check_arrays(x, y):
        return np.nan
    x, y = _prep_arrays(x, y)
    if _check_arrays(x, y):
        return np.nan
    n_sq = len(x) ** 2 + 1
    rank_x_inc = (stats.rankdata(x) ** 2) / n_sq - 0.5
    rank_y_inc = (stats.rankdata(y) ** 2) / n_sq - 0.5
    rank_y_dec = -(stats.rankdata(-y) ** 2) / n_sq + 0.5
    lower = np.sum((-rank_x_inc < rank_y_inc) * (rank_x_inc + rank_y_inc) ** 2)
    higher = np.sum((rank_x_inc > rank_y_dec) * (rank_x_inc - rank_y_dec) ** 2)
    return float((lower - higher) / (lower + higher))


def tukey_correlation(x: np.ndarray, y: np.ndarray) -> float:
    """Calculate Tukey's correlation coefficient.

    It is not quite as robust as rank correlations, but it is more
    efficient in Gaussian and near-Gaussian cases.

    The arrays will be flatten before any calculations.

    Parameters
    ----------
    x : array_like
        Input array.
    y : array_like
        Input array.

    Returns
    -------
    tcc : float.
        The value of the Tukey's correlation coefficient.

    References
    ----------
    Wainer, H.; Thissen, D. (1976).
    Three steps toward robust regression.
    Psychometrika, 41, 9-34.

    Notes
    -----
    This measure is assymetric: (x, y) != (y, x).
    """
    if _check_arrays(x, y):
        return np.nan
    x, y = _prep_arrays(x, y)
    if _check_arrays(x, y):
        return np.nan
    s_x = gini_mean_difference(x)
    s_y = gini_mean_difference(y)
    x_norm = x / s_x
    y_norm = y / s_y
    coef = 0.25 * (
        gini_mean_difference(x_norm + y_norm) ** 2
        - gini_mean_difference(x_norm - y_norm) ** 2
    )
    return float(max(min(coef, 1.0), -1.0))


def gaussain_rank_correlation(x: np.ndarray, y: np.ndarray) -> float:
    """Calculate Gaussian rank correlation coefficient.

    The Gaussian rank correlation equals the usual correlation coefficient
    computed from the normal scores of the data.

    The arrays will be flatten before any calculations.

    Parameters
    ----------
    x : array_like
        Input array.
    y : array_like
        Input array.

    Returns
    -------
    grc : float.
        The value of the Gaussian rank correlation coefficient.

    References
    ----------
    Boudt, K.; Cornelissen, J.; & Croux, C. (2012).
    The Gaussian rank correlation estimator: robustness properties.
    Statistics and Computing, 22, 471-483.
    """
    if _check_arrays(x, y):
        return np.nan
    x, y = _prep_arrays(x, y)
    if _check_arrays(x, y):
        return np.nan
    n = len(x)
    norm_factor = 1 / (n + 1)
    x_ranks_norm = stats.rankdata(x) * norm_factor
    y_ranks_norm = stats.rankdata(y) * norm_factor
    coef = np.sum(stats.norm.ppf(x_ranks_norm) * stats.norm.ppf(y_ranks_norm)) / np.sum(
        stats.norm.ppf(np.arange(1, n + 1) * norm_factor) ** 2
    )
    return float(max(min(coef, 1.0), -1.0))


def quantile_correlation(x: np.ndarray, y: np.ndarray, q: float = 0.5) -> float:
    """Calculate quantile correlation.

    This function measures linear association between two
    variables for a given quantile.

    The arrays will be flatten before any calculations.
    This implementation is dependent on the order of the arrays,
    which could be useful in application like time series.

    Parameters
    ----------
    x : array_like
        Input array.
    y : array_like
        Input array.
    q : float, default = 0.5
        Quantile used in the calculations.

    Returns
    -------
    qcor : float.
        The value of the quantile sample correlation.

    References
    ----------
    Li, G.; Li, Y.; & Tsai, C. (2015).
    Quantile Correlations and Quantile Autoregressive Modeling.
    Journal of the American Statistical Association, 110(509), 246-261.

    Notes
    -----
    This measure is assymetric: (x, y) != (y, x).
    """
    if _check_arrays(x, y):
        return np.nan
    x, y = _prep_arrays(x, y)
    if _check_arrays(x, y):
        return np.nan
    return float(
        np.mean((q - (y < np.quantile(y, q=q))) * (x - np.mean(x)))
        / (((q - q**2) * np.var(x)) ** 0.5)
    )


def normalized_chatterjee_xi(x: np.ndarray, y: np.ndarray) -> float:
    """Calculate normalizd Xi correlation coefficient.

    Another variation of rank correlation which does not make any assumptions about
    underlying distributions of the variable.

    It ranges from 0 (variables are completely independent) to 1
    (one is a measurable function of the other). This variant normalizes Chatterjee Xi,
    so it's maximum will always be 1.0.

    This implementation does not break ties at random, instead
    it break ties depending on order. This makes it dependent on
    data sorting, which could be useful in application like time
    series.

    The arrays will be flatten before any calculations.

    Parameters
    ----------
    x : array_like
        Input array.
    y : array_like
        Input array.

    Returns
    -------
    nxi : float.
        The value of the normalized xi correlation coefficient.

    References
    ----------
    Dalitz, C.; Arning, J.; Goebbels, S. (2024).
    A Simple Bias Reduction for Chatterjee's Correlation.
    J Stat Theory Pract 18, 51.

    Notes
    -----
    This measure is assymetric: (x, y) != (y, x).
    """
    if _check_arrays(x, y):
        return np.nan
    x, y = _prep_arrays(x, y)
    if _check_arrays(x, y):
        return np.nan
    n = len(x)
    # y ~ f(x)
    y_forward_ordered = y[np.argsort(x)]
    _, y_unique_indexes, y_counts = np.unique(
        y_forward_ordered, return_inverse=True, return_counts=True
    )
    right_xy = np.cumsum(y_counts)[y_unique_indexes]
    left_xy = np.cumsum(y_counts[::-1])[len(y_counts) - y_unique_indexes - 1]
    # y ~ f(y)
    y_ordered = y[np.argsort(y)]
    _, y_unique_indexes, y_counts = np.unique(
        y_ordered, return_inverse=True, return_counts=True
    )
    right_yy = np.cumsum(y_counts)[y_unique_indexes]
    left_yy = np.cumsum(y_counts[::-1])[len(y_counts) - y_unique_indexes - 1]
    # divide one by another
    return float(
        max(
            -1,
            (
                1
                - 0.5
                * np.sum(np.abs(np.diff(right_xy)))
                / np.mean(left_xy * (n - left_xy))
            )
            / (
                1
                - 0.5
                * np.sum(np.abs(np.diff(right_yy)))
                / np.mean(left_yy * (n - left_yy)),
            ),
        )
    )



"""Module for measures of central tendency."""

import math

import numpy as np
from scipy import stats  # type: ignore[import-untyped]


def midrange(x: np.ndarray) -> float:
    """Calculate midrange or midpoint, i.e. average between min and max.

    This measure could be noisy since it is based on minimum and maximum.

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    mr : float
        The value of the midrange.

    References
    ----------
    Dodge, Y. (2003).
    The Oxford dictionary of Statistical Terms.
    Oxford University Press.
    """
    maximum = np.nanmax(x)
    minimum = np.nanmin(x)
    return float((maximum + minimum) * 0.5)


def midhinge(x: np.ndarray) -> float:
    """Calculate midhinge, i.e. average between 1st and 3rd quartile.

    This measure is more robust then average.

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    mh : float
        The value of the midhinge.

    References
    ----------
    Tukey, J. W. (1977).
    Exploratory Data Analysis.
    Addison-Wesley.
    """
    q1, q3 = np.nanquantile(x, [0.25, 0.75])
    return float((q3 + q1) * 0.5)


def trimean(x: np.ndarray) -> float:
    """Calculate trimean, i.e weighted average between 3 quartiles.

    This measure is more robust then average.

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    tm : float
        The value of the trimean.

    References
    ----------
    Tukey, J. W. (1977).
    Exploratory Data Analysis.
    Addison-Wesley.
    """
    q1, q2, q3 = np.nanquantile(x, [0.25, 0.5, 0.75])
    return float(0.5 * q2 + 0.25 * q1 + 0.25 * q3)


def contraharmonic_mean(x: np.ndarray) -> float:
    """Calculate contraharmonic mean.

    Contraharmonic mean is a function complementary to the harmonic mean.
    The contraharmonic mean is a special case of the Lehmer mean with p=2.
    Mostly used in signal processing (for example filters).

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    chm : float
        The value of the contraharmonic mean.

    References
    ----------
    Bullen, P. S. (1987).
    Handbook of means and their inequalities.
    Springer.
    """
    return float(np.nansum(np.square(x)) / np.nansum(x))


def midmean(x: np.ndarray) -> float:
    """Calculate interquartile mean, i.e mean inside interquartile range.

    This measure is more robust then average.

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    iqm : float
        The value of the interquartile mean.

    References
    ----------
    Salkind, N. J. (2008).
    Encyclopedia of Research Design.
    SAGE Publications, Inc.
    """
    q1, q3 = np.nanquantile(x, [0.25, 0.75])
    return float(np.nanmean(np.where((x >= q1) & (x <= q3), x, np.nan)))


def hodges_lehmann_sen_location(x: np.ndarray) -> float:
    """Calculate Hodges-Lehmann-Sen robust location measure (pseudomedian).

    This measure is more robust then average.

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    hls : float
        The value of Hodges-Lehmann-Sen estimator.

    References
    ----------
    Hodges, J. L.; Lehmann, E. L. (1963).
    Estimation of location based on ranks.
    Annals of Mathematical Statistics. 34 (2): 598-611.

    Sen, P. K. (1963).
    On the Estimation of Relative Potency in Dilution (-Direct)
    Assays by Distribution-Free Methods.
    Biometrics 19, no. 4: 532-552.

    Notes
    -----
    This implementation uses cartesian product, so the time and memory complexity
    are N^2. It is best to not use it on large arrays.
    """
    # In the original paper authors suggest use only upper triangular
    # of the cartesian product, but in this implementation we use
    # whole matrix, which is equvalent.
    product = np.meshgrid(x, x, sparse=True)
    return float(np.nanmedian(product[0] + product[1]) * 0.5)


def standard_trimmed_harrell_davis_quantile(x: np.ndarray, q: float = 0.5) -> float:
    """Calculate Standard Trimmed Harrell-Davis median estimator.

    This measure is very robust.
    It calculates weighted Harrel-Davies quantiles on only sqrt(N) samples,
    located in the region with the most probability mass.

    Parameters
    ----------
    x : array_like
        Input array.
    q : float, default = 0.5
        Quantile value in range (0, 1).

    Returns
    -------
    thdq : float
        The value of Trimmed Harrell-Davis quantile.

    References
    ----------
    Akinshin, A. (2022).
    Trimmed Harrell-Davis quantile estimator based on
    the highest density interval of the given width.
    Communications in Statistics - Simulation and Computation, pp. 1-11.

    See Also
    --------
    scipy.stats.mstats.hdquantiles - Harrell-Davis quantile estimates.
    """
    if q <= 0 or q >= 1:
        msg = "Parameter q should be in range (0, 1)."
        raise ValueError(msg)
    _x = np.sort(x)
    _x = _x[np.isfinite(_x)]
    n = len(_x)
    if n == 0:
        return np.nan
    if n == 1:
        return float(_x[0])
    n_calculated = 1 / n**0.5  # heuristic suggested by the author
    a = (n + 1) * q
    b = (n + 1) * (1.0 - q)
    hdi = (max(0, q - n_calculated * 0.5), min(1, q + n_calculated * 0.5))
    hdi_cdf = stats.beta.cdf(hdi, a, b)
    i_start = math.floor(hdi[0] * n)
    i_end = math.ceil(hdi[1] * n)
    nums = np.arange(i_start, i_end + 1) / n
    nums[nums <= hdi[0]] = hdi[0]
    nums[nums >= hdi[1]] = hdi[1]
    cdfs = (stats.beta.cdf(nums, a, b) - hdi_cdf[0]) / (hdi_cdf[1] - hdi_cdf[0])
    w = cdfs[1:] - cdfs[:-1]
    return float(np.sum(_x[i_start:i_end] * w))


def half_sample_mode(x: np.ndarray) -> float:
    """Calculate half sample mode.

    This estimator is more stable than regular mode estimation,
    especially for floating point values.

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    hsm : float
        The value of half sample mode.

    References
    ----------
    Robertson, T.; Cryer, J. D. (1974).
    An Iterative Procedure for Estimating the Mode.
    Journal of the American Statistical Association, 69(348), 1012-1016.

    See Also
    --------
    scipy.stats.mode - Mode estimator.
    """
    # heavily inspired by https://github.com/cran/modeest/blob/master/R/hsm.R
    y = np.sort(x)
    y = y[np.isfinite(y)]
    _corner_cases = (4, 3)  # for 4 samples and 3 samples
    while (ny := len(y)) >= _corner_cases[0]:
        half_y = math.ceil(ny / 2)
        w_min = y[-1] - y[0]
        for i in range(ny - half_y):
            w = y[i + half_y - 1] - y[i]
            if w <= w_min:
                w_min = w
                j = i
        if w == 0:
            return float(y[j])
        y = y[j : (j + half_y - 1)]
    if len(y) == _corner_cases[1]:
        z = 2 * y[1] - y[0] - y[2]
        if z < 0:
            return float(np.mean(y[0:1]))
        if z > 0:
            return float(np.mean(y[1:2]))
        return float(y[1])
    return float(np.mean(y))


def tau_location(x: np.ndarray, c: float = 4.5) -> float:
    """Calculate Tau measure of location.

    This measure is very robust and has higher efficiency than median.

    Parameters
    ----------
    x : array_like
        Input array.
    c : float, default = 4.5
        Constant that filter outliers.

    Returns
    -------
    tml : float
        The value of Tau location.

    References
    ----------
    Wilcox, R. (2012).
    Introduction to Robust Estimation Hypothesis Testing.
    3rd Edition, Academic Press, New York.
    """
    if c <= 0:
        msg = "Parameter c should be strictly positive."
        raise ValueError(msg)
    med = np.nanmedian(x)
    mad = np.nanmedian(np.abs(x - med))
    scaled_x = (x - med) / mad
    w = np.square(1.0 - np.square(scaled_x / c)) * (np.abs(scaled_x) <= c)
    return float(np.nansum(x * w) / np.nansum(w))


def grenanders_m(x: np.ndarray, p: float = 1.001, k: int = 2) -> float:
    """Calculate Grenander's Mode.

    This measure is a direct nonparametric estimation of the mode.
    It is not very robust to outliers.
    It is recommended to set k > 2p, but it is only possible if x has
    enough samples.

    Parameters
    ----------
    x : array_like
        Input array.
    p : float, default = 1.001
        Smoothing constant.
    k : int, default = 2
        The number of samples to filter.

    Returns
    -------
    gm : float
        The value of Grenander's Mode

    References
    ----------
    Grenander, U. (1965).
    Some direct estimates of the mode.
    Annals of Mathematical Statistics, 36, 131-138.
    """
    x_sort = np.sort(x)
    x_sort = x_sort.astype("float")
    x_sort = x_sort[np.isfinite(x_sort)]

    if p <= 1:
        msg = "Parameter p should be a float greater than 1."
        raise ValueError(msg)
    if (k <= 0) or (not isinstance(k, int)):
        msg = "Parameter k should be an integer between 1 and lenght of x."
        raise ValueError(msg)

    if len(x_sort) <= k:
        return np.nan

    # pre calculate diffs
    diff = x_sort[k:] - x_sort[:-k]
    # if the diffs are constant - return the value
    if diff.sum() == 0.0:
        return float(x_sort[0])
    # to avoid division by zero
    diff[diff == 0.0] = np.nan

    return float(
        0.5
        * np.nansum((x_sort[k:] + x_sort[:-k]) / np.power(diff, p))
        / np.nansum(np.power(diff, -p))
    )


def gastwirth_location(x: np.ndarray) -> float:
    """Calculate Gastwirth's location estimator.

    This measure is more robust then average.

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    gle : float
        The value of the Gastwirth's location.

    References
    ----------
    Gastwirth, J. L. (1966).
    On Robust Procedures.
    J. Amer. Statist. Assn., Vol. 61, pp. 929-948.
    """
    p33, p50, p66 = np.nanquantile(x, [1 / 3, 0.5, 2 / 3])
    return float(0.3 * p33 + 0.4 * p50 + 0.3 * p66)
